@@ -20,10 +20,10 @@
 #
 #
 
-from casadi import integrator, Function, MX, hcat, vertcat, vcat, linspace, veccat, DM, repmat, horzsplit, cumsum, inf, mtimes, symvar, horzcat, symvar, vvcat
+from casadi import integrator, Function, MX, hcat, vertcat, vcat, linspace, veccat, DM, repmat, horzsplit, cumsum, inf, mtimes, symvar, horzcat, symvar, vvcat, is_equal
 from .direct_method import DirectMethod
 from .splines import BSplineBasis, BSpline
-from .casadi_helpers import reinterpret_expr
+from .casadi_helpers import reinterpret_expr, HashOrderedDict
 from numpy import nan, inf
 import numpy as np
 from collections import defaultdict
@@ -43,8 +43,8 @@ class Grid:
         pass
 
 class FixedGrid(Grid):
-    def __init__(self, localize_t0=False, localize_T=False, *args, **kwargs):
-        Grid.__init__(self, *args, **kwargs)
+    def __init__(self, localize_t0=False, localize_T=False, **kwargs):
+        Grid.__init__(self, **kwargs)
         self.localize_t0 = localize_t0
         self.localize_T = localize_T
 
@@ -90,8 +90,8 @@ class FreeGrid(FixedGrid):
         Enforced with constraints
         Default: inf
     """
-    def __init__(self, *args, **kwargs):
-        FixedGrid.__init__(self, *args, **kwargs)
+    def __init__(self, **kwargs):
+        FixedGrid.__init__(self, **kwargs)
         self.localize_T = True
 
     def constrain_T(self,opti,T,Tnext,N):
@@ -121,8 +121,8 @@ class UniformGrid(FixedGrid):
         Enforced with constraints
         Default: inf
     """
-    def __init__(self, *args, **kwargs):
-        FixedGrid.__init__(self, *args, **kwargs)
+    def __init__(self, **kwargs):
+        FixedGrid.__init__(self, **kwargs)
 
     def constrain_T(self,opti,T,Tnext,N):
         opti.subject_to(T==Tnext)
@@ -169,18 +169,18 @@ class GeometricGrid(FixedGrid):
     >>> MultipleShooting(N=3, grid=GeometricGrid(2)) # grid = [0, 1, 3, 7]*T/7
     >>> MultipleShooting(N=3, grid=GeometricGrid(2,local=True)) # grid = [0, 1, 5, 21]*T/21
     """
-    def __init__(self, growth_factor, local=False, *args, **kwargs):
+    def __init__(self, growth_factor, local=False, **kwargs):
         assert growth_factor>=1
         self._growth_factor = growth_factor
         self.local = local
-        FixedGrid.__init__(self, *args, **kwargs)
+        FixedGrid.__init__(self, **kwargs)
 
     def __call__(self, t0, T, N):
         n = self.normalized(N)
         return t0 + hcat(n)*T
 
     def growth_factor(self, N):
-        if not self.local:
+        if not self.local and N>1:
             return self._growth_factor**(1.0/(N-1))
         return self._growth_factor
 
@@ -253,41 +253,47 @@ class SamplingMethod(DirectMethod):
         self.q = 0
 
     def discrete_system(self, stage):
-        f = stage._ode()
-
         # Coefficient matrix from RK4 to reconstruct 4th order polynomial (k1,k2,k3,k4)
         # nstates x (4 * M)
         poly_coeffs = []
+        poly_coeffs_z = []
 
         t0 = MX.sym('t0')
         T = MX.sym('T')
         DT = T / self.M
 
         # Size of integrator interval
-        X0 = f.mx_in("x")            # Initial state
-        U = f.mx_in("u")             # Control
-        P = f.mx_in("p")
-        Z = f.mx_in("z")
+        X0 = MX.sym("x", stage.nx)            # Initial state
+        U = MX.sym("u", stage.nu)             # Control
+        P = MX.sym("p", stage.np+stage.v.shape[0])
+        Z = MX.sym("z", stage.nz)
 
         X = [X0]
-        if hasattr(self, 'intg_' + self.intg):
-            intg = getattr(self, "intg_" + self.intg)(f, X0, U, P, Z)
-        else:
-            intg = self.intg_builtin(f, X0, U, P, Z)
-        assert not intg.has_free()
+        Zs = []
 
         # Compute local start time
         t0_local = t0
         quad = 0
+
+        if stage._state_next:
+            intg = stage._diffeq()
+        elif hasattr(self, 'intg_' + self.intg):
+            intg = getattr(self, "intg_" + self.intg)(stage._ode(), X0, U, P, Z)
+        else:
+            intg = self.intg_builtin(stage._ode(), X0, U, P, Z)
+        assert not intg.has_free()
+    
         for j in range(self.M):
             intg_res = intg(x0=X[-1], u=U, t0=t0_local, DT=DT, p=P)
             X.append(intg_res["xf"])
+            Zs.append(intg_res["zf"])
             quad = quad + intg_res["qf"]
             poly_coeffs.append(intg_res["poly_coeff"])
+            poly_coeffs_z.append(intg_res["poly_coeff_z"])
             t0_local += DT
 
-        ret = Function('F', [X0, U, T, t0, P], [X[-1], hcat(X), hcat(poly_coeffs), quad],
-                       ['x0', 'u', 'T', 't0', 'p'], ['xf', 'Xi', 'poly_coeff', 'qf'])
+        ret = Function('F', [X0, U, T, t0, P], [X[-1], hcat(X), hcat(poly_coeffs), quad, Zs[-1], hcat(Zs), hcat(poly_coeffs_z)],
+                       ['x0', 'u', 'T', 't0', 'p'], ['xf', 'Xi', 'poly_coeff', 'qf', 'zf', 'Zi', 'poly_coeff_z'])
         assert not ret.has_free()
         return ret
 
@@ -296,8 +302,8 @@ class SamplingMethod(DirectMethod):
         DT = MX.sym("DT")
         t0 = MX.sym("t0")
         # A single Runge-Kutta 4 step
-        k1 = f(x=X, u=U, p=P, t=t0, z=Z)
-        k2 = f(x=X + DT / 2 * k1["ode"], u=U, p=P, t=t0+DT/2, z=Z)
+        k1 = f(x=X, u=U, p=P, t=t0)
+        k2 = f(x=X + DT / 2 * k1["ode"], u=U, p=P, t=t0+DT/2)
         k3 = f(x=X + DT / 2 * k2["ode"], u=U, p=P, t=t0+DT/2)
         k4 = f(x=X + DT * k3["ode"], u=U, p=P, t=t0+DT)
 
@@ -306,7 +312,7 @@ class SamplingMethod(DirectMethod):
         f2 = 4/DT**2*(k3["ode"]-k2["ode"])/6
         f3 = 4*(k4["ode"]-2*k3["ode"]+k1["ode"])/DT**3/24
         poly_coeff = hcat([X, f0, f1, f2, f3])
-        return Function('F', [X, U, t0, DT, P], [X + DT / 6 * (k1["ode"] + 2 * k2["ode"] + 2 * k3["ode"] + k4["ode"]), poly_coeff, DT / 6 * (k1["quad"] + 2 * k2["quad"] + 2 * k3["quad"] + k4["quad"])], ['x0', 'u', 't0', 'DT', 'p'], ['xf', 'poly_coeff', 'qf'])
+        return Function('F', [X, U, t0, DT, P], [X + DT / 6 * (k1["ode"] + 2 * k2["ode"] + 2 * k3["ode"] + k4["ode"]), poly_coeff, DT / 6 * (k1["quad"] + 2 * k2["quad"] + 2 * k3["quad"] + k4["quad"]), MX(0, 1), MX()], ['x0', 'u', 't0', 'DT', 'p'], ['xf', 'poly_coeff', 'qf', 'zf', 'poly_coeff_z'])
 
     def intg_builtin(self, f, X, U, P, Z):
         # A single CVODES step
@@ -317,10 +323,12 @@ class SamplingMethod(DirectMethod):
         data = {'x': X, 'p': vertcat(U, DT, P, t0), 'z': Z, 't': t, 'ode': DT * res["ode"], 'quad': DT * res["quad"], 'alg': res["alg"]}
         options = dict(self.intg_options)
         if self.intg in ["collocation"]:
-            options["number_of_finite_elements"] = 1
-        I = integrator('intg_cvodes', self.intg, data, options)
+            # In rockit, M replaces number_of_finite_elements on a higher level
+            if "number_of_finite_elements" not in options:
+                options["number_of_finite_elements"] = 1
+        I = integrator('intg_'+self.intg, self.intg, data, options)
         res = I.call({'x0': X, 'p': vertcat(U, DT, P, t0)})
-        return Function('F', [X, U, t0, DT, P], [res["xf"], MX(), res["qf"]], ['x0', 'u', 't0', 'DT', 'p'], ['xf', 'poly_coeff','qf'])
+        return Function('F', [X, U, t0, DT, P], [res["xf"], MX(), res["qf"], res["zf"], MX()], ['x0', 'u', 't0', 'DT', 'p'], ['xf', 'poly_coeff','qf','zf','poly_coeff_z'])
 
     def transcribe_placeholders(self, stage, placeholders):
         """
@@ -328,33 +336,16 @@ class SamplingMethod(DirectMethod):
         """
         return stage._transcribe_placeholders(self, placeholders)
 
-    def transcribe(self, stage, opti):
+    def transcribe(self, stage, pass_nr=1,**kwargs):
         """
         Transcription is the process of going from a continuous-time OCP to an NLP
         """
+        if pass_nr>1: return
+        opti = stage.master._method.opti
         DM.set_precision(14)
         self.add_variables(stage, opti)
         self.add_parameter(stage, opti)
 
-        # Create time grid (might be symbolic)
-        self.T = self.eval(stage, stage._T)
-        self.t0 = self.eval(stage, stage._t0)
-
-        self.set_initial(stage, opti, stage._initial)
-        T_init = opti.debug.value(self.T, opti.initial())
-        t0_init = opti.debug.value(self.t0, opti.initial())
-
-        # How to get initial value -> ask opti?
-
-        control_grid_init = self.time_grid(t0_init, T_init, self.N)
-        if self.time_grid.localize_t0:
-            for k in range(1, self.N):
-                stage.set_initial(self.t0_local[k], control_grid_init[k])
-            stage.set_initial(self.t0_local[self.N], control_grid_init[self.N])
-        if self.time_grid.localize_T:
-            for k in range(not isinstance(self.time_grid, FreeGrid), self.N):
-                stage.set_initial(self.T_local[k], control_grid_init[k+1]-control_grid_init[k])
-
         self.integrator_grid = []
         for k in range(self.N):
             t_local = linspace(self.control_grid[k], self.control_grid[k+1], self.M+1)
@@ -363,9 +354,28 @@ class SamplingMethod(DirectMethod):
         self.add_constraints(stage, opti)
         self.add_constraints_after(stage, opti)
         self.add_objective(stage, opti)
-        self.set_initial(stage, opti, stage._initial)
-        self.set_parameter(stage, opti)
+    
+        # Create time grid (might be symbolic)
+        self.T = self.eval(stage, stage._T)
+        self.t0 = self.eval(stage, stage._t0)
 
+        self.set_initial(stage, opti, stage._initial)
+        T_init = opti.debug.value(self.T, opti.initial())
+        t0_init = opti.debug.value(self.t0, opti.initial())
+
+        initial = HashOrderedDict()
+        # How to get initial value -> ask opti?
+        control_grid_init = self.time_grid(t0_init, T_init, self.N)
+        if self.time_grid.localize_t0:
+            for k in range(1, self.N):
+                initial[self.t0_local[k]] = control_grid_init[k]
+            initial[self.t0_local[self.N]] = control_grid_init[self.N]
+        if self.time_grid.localize_T:
+            for k in range(not isinstance(self.time_grid, FreeGrid), self.N):
+                initial[self.T_local[k]] = control_grid_init[k+1]-control_grid_init[k]
+
+        self.set_initial(stage, opti, initial)
+        self.set_parameter(stage, opti)
 
     def add_constraints_before(self, stage, opti):
         for c, meta, _ in stage._constraints["point"]:
@@ -401,7 +411,8 @@ class SamplingMethod(DirectMethod):
         lookup = dict(zip(subst_from, subst_to))
 
         subst_from += stage._inf_der.keys()
-        subst_to += [lookup[e].derivative() for e in stage._inf_der.values()]
+        dt = (self.control_grid[k + 1] - self.control_grid[k])/self.M
+        subst_to += [lookup[e].derivative()*(1/dt) for e in stage._inf_der.values()]
 
         subst_from += stage._inf_inert.keys()
         subst_to += stage._inf_inert.values()
@@ -442,10 +453,7 @@ class SamplingMethod(DirectMethod):
         opti.add_objective(self.eval(stage, stage._objective))
 
     def add_variables_V(self, stage, opti):
-        V = []
-        for v in stage.variables['']:
-            V.append(opti.variable(v.shape[0], v.shape[1]))
-        self.V = veccat(*V)
+        DirectMethod.add_variables(self, stage, opti)
 
         # Create time grid (might be symbolic)
         self.T = self.eval(stage, stage._T)
@@ -487,7 +495,7 @@ class SamplingMethod(DirectMethod):
         self.time_grid.bounds_T(opti, self.T_local, self.t0_local, k, self.T, self.N)
 
     def get_p_control_at(self, stage, k=-1):
-        return veccat(*[p[:,k] for p in self.P_control])
+        return veccat(*[p[k] for p in self.P_control])
 
     def get_v_control_at(self, stage, k=-1):
         return veccat(*[v[k] for v in self.V_control])
@@ -499,7 +507,7 @@ class SamplingMethod(DirectMethod):
         return vertcat(vvcat(self.P), self.get_p_control_at(stage, k), self.V, self.get_v_control_at(stage, k))
 
     def eval(self, stage, expr):
-        return stage._expr_apply(expr, p=veccat(*self.P), v=self.V, t0=self.t0, T=self.T)
+        return stage.master._method.eval_top(stage.master, stage._expr_apply(expr, p=veccat(*self.P), v=self.V, t0=self.t0, T=self.T))
 
     def eval_at_control(self, stage, expr, k):
         try:
@@ -524,9 +532,21 @@ class SamplingMethod(DirectMethod):
             subst_from.append(vvcat(symbols[offset]))
             subst_to.append(self._eval_at_control(stage, vvcat(offsets[offset]), k+offset))
             #print(expr, subst_from, subst_to)
-        return stage._expr_apply(expr, sub=(subst_from, subst_to), t0=self.t0, T=self.T, x=self.X[k], z=self.Z[k] if self.Z else nan, xq=self.q if k==-1 else nan, u=self.U[k], p_control=self.get_p_control_at(stage, k), v=self.V, p=veccat(*self.P), v_control=self.get_v_control_at(stage, k),  v_states=self.get_v_states_at(stage, k), t=self.control_grid[k])
+
+        expr = stage._expr_apply(expr, sub=(subst_from, subst_to), t0=self.t0, T=self.T, x=self.X[k], z=self.Z[k] if self.Z else nan, xq=self.q if k==-1 else nan, u=self.U[k], p_control=self.get_p_control_at(stage, k), v=self.V, p=veccat(*self.P), v_control=self.get_v_control_at(stage, k),  v_states=self.get_v_states_at(stage, k), t=self.control_grid[k])
+        expr = stage.master._method.eval_top(stage.master, expr)
+        #print("expr",expr)
+        return expr
 
     def _eval_at_control(self, stage, expr, k):
+        x = self.X[k]
+        z = self.Z[k] if self.Z else nan
+        u = self.U[-1] if k==len(self.U) else self.U[k] # Would be fixed if we remove the (k=-1 case)
+        p_control = self.get_p_control_at(stage, k)
+        v_control = self.get_v_control_at(stage, k)
+        v_states = self.get_v_states_at(stage, k)
+        t = self.control_grid[k]
+        return stage._expr_apply(expr, t0=self.t0, T=self.T, x=x, z=z, xq=self.q if k==-1 else nan, u=u, p_control=p_control, v=self.V, p=veccat(*self.P), v_control=v_control,  v_states=v_states, t=t)
         kwargs = dict(t0=self.t0, T=self.T, v=self.V, p=veccat(*self.P))
         try:
             kwargs["x"] = self.X[k]
@@ -541,7 +561,7 @@ class SamplingMethod(DirectMethod):
         except IndexError:
             pass
         try:
-            kwargs["u"] = self.control_grid[k]
+            kwargs["u"] = self.U[k]
         except IndexError:
             pass
         try:
@@ -563,17 +583,18 @@ class SamplingMethod(DirectMethod):
         return stage._expr_apply(expr, **kwargs)
 
     def eval_at_integrator(self, stage, expr, k, i):
-        return stage._expr_apply(expr, t0=self.t0, T=self.T, x=self.xk[k*self.M + i], z=self.zk[k*self.M + i] if self.zk else nan, u=self.U[k], p_control=self.get_p_control_at(stage, k), v=self.V, p=veccat(*self.P), v_control=self.get_v_control_at(stage, k),  v_states=self.get_v_states_at(stage, k), t=self.integrator_grid[k][i])
+        return stage.master._method.eval_top(stage.master, stage._expr_apply(expr, t0=self.t0, T=self.T, x=self.xk[k*self.M + i], z=self.zk[k*self.M + i] if self.zk else nan, u=self.U[k], p_control=self.get_p_control_at(stage, k), v=self.V, p=veccat(*self.P), v_control=self.get_v_control_at(stage, k),  v_states=self.get_v_states_at(stage, k), t=self.integrator_grid[k][i]))
 
     def eval_at_integrator_root(self, stage, expr, k, i, j):
-        return stage._expr_apply(expr, t0=self.t0, T=self.T, x=self.xr[k][i][:,j], z=self.zr[k][i][:,j] if self.zk else nan, u=self.U[k], p_control=self.get_p_control_at(stage, k), v=self.V, p=veccat(*self.P), v_control=self.get_v_control_at(stage, k), t=self.tr[k][i][j])
+        return stage.master._method.eval_top(stage.master, stage._expr_apply(expr, t0=self.t0, T=self.T, x=self.xr[k][i][:,j], z=self.zr[k][i][:,j] if self.zk else nan, u=self.U[k], p_control=self.get_p_control_at(stage, k), v=self.V, p=veccat(*self.P), v_control=self.get_v_control_at(stage, k), t=self.tr[k][i][j]))
 
-    def set_initial(self, stage, opti, initial):
+    def set_initial(self, stage, master, initial):
+        opti = master.opti if hasattr(master, 'opti') else master
         opti.cache_advanced()
         for var, expr in initial.items():
-            if var is stage.T:
+            if is_equal(var, stage.T):
                 var = self.T
-            if var is stage.t0:
+            if is_equal(var, stage.t0):
                 var = self.t0
             opti_initial = opti.initial()
             for k in list(range(self.N))+[-1]:
@@ -592,22 +613,27 @@ class SamplingMethod(DirectMethod):
                         value = value[:,k]
                 opti.set_initial(target, value, cache_advanced=True)
 
-    def set_value(self, stage, opti, parameter, value):
+    def set_value(self, stage, master, parameter, value):
+        opti = master.opti if hasattr(master, 'opti') else master
+        found = False
         for i, p in enumerate(stage.parameters['']):
-            if parameter is p:
+            if is_equal(parameter, p):
+                found = True
                 opti.set_value(self.P[i], value)
         for i, p in enumerate(stage.parameters['control']):
-            if parameter is p:
-                opti.set_value(self.P_control[i], value)
+            if is_equal(parameter, p):
+                found = True
+                opti.set_value(hcat(self.P_control[i]), value)
+        assert found, "Parameter not found"
 
     def add_parameter(self, stage, opti):
         for p in stage.parameters['']:
             self.P.append(opti.parameter(p.shape[0], p.shape[1]))
         for p in stage.parameters['control']:
-            self.P_control.append(hcat([opti.parameter(p.shape[0], p.shape[1]) for i in range(self.N)]))
+            self.P_control.append([opti.parameter(p.shape[0], p.shape[1]) for i in range(self.N)])
 
     def set_parameter(self, stage, opti):
         for i, p in enumerate(stage.parameters['']):
             opti.set_value(self.P[i], stage._param_vals[p])
         for i, p in enumerate(stage.parameters['control']):
-            opti.set_value(self.P_control[i], stage._param_vals[p])
+            opti.set_value(hcat(self.P_control[i]), stage._param_vals[p])
